Imports System.Device.Location
Imports Microsoft.Devices.Sensors
Imports Petzold.Astronomy

Namespace AstroPhone
	Public Class Game1
		Inherits Microsoft.Xna.Framework.Game

		Private Const WIDTH As Integer = 480
		Private Const HEIGHT As Integer = 800
		Private Const PIXELS_PER_DEGREE As Integer = 20

		Private graphics As GraphicsDeviceManager
		Private spriteBatch As SpriteBatch

		Private labelFont As SpriteFont
		Private starDotRenderer As DotRenderer
		Private planetDotRenderer As DotRenderer
		Private lineRenderer As LineRenderer
		Private displayMatrix As Matrix = Matrix.Identity

		Private motion As Motion
        Private geoWatcher As GeoCoordinateWatcher

		Private coordinateProjection As New HorizontalCoordinateProjection

		Private constellations As Constellations
		Private visibleStars As New List(Of Star)
		Private systemBodies As New List(Of SolarSystemBody)
		Private celestialBodies As New List(Of CelestialBody)
		Private updateIndex As Integer

		' Points for horizon line
		Private horizon(23) As Vector2
		Private horizonLabels() As String = { "N", "NE", "E", "SE", "S", "SW", "W", "NW" }

		' Points for ecliptic line
		Private ecliptic(23) As Vector2
		Private eclipticTicks(23, 1) As Vector2

		Public Sub New()
			graphics = New GraphicsDeviceManager(Me)
			Content.RootDirectory = "Content"

			' Frame rate is 30 fps by default for Windows Phone.
			TargetElapsedTime = TimeSpan.FromTicks(333333)

			' Extend battery life under lock.
			InactiveSleepTime = TimeSpan.FromSeconds(1)

			' Switch to portrait mode
			graphics.IsFullScreen = True
			graphics.PreferredBackBufferWidth = WIDTH
			graphics.PreferredBackBufferHeight = HEIGHT

			' Create Motion object
			If Motion.IsSupported Then
				motion = New Motion With {.TimeBetweenUpdates = Me.TargetElapsedTime}
			End If

			geoWatcher = New GeoCoordinateWatcher
		End Sub

		Protected Overrides Sub OnActivated(ByVal sender As Object, ByVal args As EventArgs)
			If motion IsNot Nothing Then
				Try
					motion.Start()
				Catch
				End Try
			End If

			geoWatcher.Start()

			' Check if data needs to be recreated
			If constellations Is Nothing Then
				constellations = Constellations.Load

				For Each constellation As Constellation In constellations.ConstellationList
					celestialBodies.Add(constellation)

					For Each star As Star In constellation.Stars
						If star.Magnitude < 2 OrElse star.IsConnectedInConstellation Then
							visibleStars.Add(star)
							celestialBodies.Add(star)
						End If
					Next star

					If constellation.Connectors IsNot Nothing AndAlso constellation.Connectors.Count > 0 Then
						constellation.StarConnectors = New List(Of StarConnector)

						' Get the actual Star objects to make the connectors more efficient
						For Each connector As Connector In constellation.Connectors
							constellation.StarConnectors.Add(New StarConnector With {.From = constellation.FindStar(connector.From), .To = constellation.FindStar(connector.To)})
						Next connector
					End If
				Next constellation

				systemBodies.Add(New Sun)
				systemBodies.Add(New Planet("mer", "Mercury", Color.Gray))
				systemBodies.Add(New Planet("ven", "Venus", Color.Yellow))
				systemBodies.Add(New Planet("mar", "Mars", Color.Red))
				systemBodies.Add(New Planet("jup", "Jupiter", Color.Orange))
				systemBodies.Add(New Planet("sat", "Saturn", Color.Brown))
				systemBodies.Add(New Planet("ura", "Uranus", Color.Green))
				systemBodies.Add(New Planet("nep", "Neptune", Color.Blue))

				' Add Earth to celestialBodies for updating but not display
				celestialBodies.Add(Earth.Instance)

				' Add Sun and planets to celestialBodies
				For Each systemBody As SolarSystemBody In systemBodies
					celestialBodies.Add(systemBody)
				Next systemBody

				' Initialize time
				Dim time As Time = Time.GetNow

				' Initialize geographic location
				Dim location As GeoCoordinate = geoWatcher.Position.Location
				Dim geographicCoordinate As New GeographicCoordinate(Angle.FromDegrees(-location.Longitude), Angle.FromDegrees(location.Latitude))

				' Initialize all celestial bodies
				For Each celestialBody As CelestialBody In celestialBodies
					celestialBody.Update(time, geographicCoordinate)
				Next celestialBody
			End If
			MyBase.OnActivated(sender, args)
		End Sub

		Protected Overrides Sub OnDeactivated(ByVal sender As Object, ByVal args As EventArgs)
			If motion IsNot Nothing Then
				motion.Stop()
			End If

			geoWatcher.Stop()
			MyBase.OnDeactivated(sender, args)
		End Sub

		Protected Overrides Sub LoadContent()
			' Create a new SpriteBatch, which can be used to draw textures.
			spriteBatch = New SpriteBatch(GraphicsDevice)

			' Create font, dot and line renderers
			labelFont = Me.Content.Load(Of SpriteFont)("LabelFont")
			starDotRenderer = New DotRenderer(Me.GraphicsDevice, 12)
			planetDotRenderer = New DotRenderer(Me.GraphicsDevice, 24)
			lineRenderer = New LineRenderer(Me.GraphicsDevice)
		End Sub

		Protected Overrides Sub Update(ByVal gameTime As GameTime)
			' Allows the game to exit
			If GamePad.GetState(PlayerIndex.One).Buttons.Back = ButtonState.Pressed Then
				Me.Exit()
			End If

			If celestialBodies.Count > 0 AndAlso motion IsNot Nothing AndAlso motion.IsDataValid Then
				' Get new time and geographic location
				Dim time As Time = Time.GetNow
				Dim location As GeoCoordinate = geoWatcher.Position.Location
				Dim geographicCoordinate As New GeographicCoordinate(Angle.FromDegrees(-location.Longitude), Angle.FromDegrees(location.Latitude))

				' Only update one celestial body at a time
				celestialBodies(updateIndex).Update(time, geographicCoordinate)
				updateIndex = (updateIndex + 1) Mod celestialBodies.Count

				' Get the current motion matrix
				Dim matrix As Matrix = motion.CurrentValue.Attitude.RotationMatrix
				Dim viewCoord As HorizontalCoordinate = HorizontalCoordinate.From(matrix)

				' Calculate the global rotation
				displayMatrix = Matrix.CreateTranslation(-WIDTH \ 2, -HEIGHT \ 2, 0) * Matrix.CreateRotationZ(CSng(-viewCoord.Tilt.Radians)) * Matrix.CreateTranslation(WIDTH \ 2, HEIGHT \ 2, 0)

				' Calculate screen points for all celestial bodies
				coordinateProjection.SetViewCenter(viewCoord)

				For Each celestialBody As CelestialBody In celestialBodies
					Dim point As Vector2 = CalculatePoint(celestialBody.HorizontalCoordinate)
					celestialBody.ScreenX = point.X
					celestialBody.ScreenY = point.Y
				Next celestialBody

				' Calculate points for horizon line
				For azimuthIndex  = 0 To horizon.Length - 1
					Dim azimuth As Angle = azimuthIndex * Angle.Circle / horizon.Length
					horizon(azimuthIndex) = CalculatePoint(New HorizontalCoordinate(azimuth, Angle.Zero))
				Next azimuthIndex

				' Calculate points for ecliptic line
				For longitudeIndex  = 0 To ecliptic.Length - 1
					Dim longitude As Angle = longitudeIndex * Angle.Circle / ecliptic.Length
					ecliptic(longitudeIndex) = CalculateEclipticPoint(longitude, Angle.Zero, time, geographicCoordinate)

					' Tick marks
					eclipticTicks(longitudeIndex, 0) = CalculateEclipticPoint(longitude, Angle.FromDegrees(1), time, geographicCoordinate)
					eclipticTicks(longitudeIndex, 1) = CalculateEclipticPoint(longitude, Angle.FromDegrees(-1), time, geographicCoordinate)
				Next longitudeIndex
			End If

			MyBase.Update(gameTime)
		End Sub

		Private Function CalculateEclipticPoint(ByVal longitude As Angle, ByVal latitude As Angle, ByVal time As Time, ByVal geographicCoordinate As GeographicCoordinate) As Vector2
			Dim eclipticCoordinate As New EclipticCoordinate(longitude, latitude, 0)
			Dim equatorialCoordinate As EquatorialCoordinate = EquatorialCoordinate.From(eclipticCoordinate, time)
			Dim horizontalCoordinate As HorizontalCoordinate = HorizontalCoordinate.From(equatorialCoordinate, geographicCoordinate, time)
			Return CalculatePoint(horizontalCoordinate)
		End Function

		Private Function CalculatePoint(ByVal horizontalCoordinate As HorizontalCoordinate) As Vector2
			Dim horzAngle, vertAngle As Angle
			coordinateProjection.GetAngleOffsets(horizontalCoordinate, horzAngle, vertAngle)
			Dim x As Single = Single.NaN
			Dim y As Single = Single.NaN

			If horzAngle > -Angle.Right AndAlso horzAngle < Angle.Right AndAlso vertAngle > -Angle.Right AndAlso vertAngle < Angle.Right Then
				x = CSng(WIDTH \ 2 + PIXELS_PER_DEGREE * horzAngle.Degrees)
				y = CSng(HEIGHT \ 2 + PIXELS_PER_DEGREE * vertAngle.Degrees)
			End If
			Return New Vector2(x, y)
		End Function

		Protected Overrides Sub Draw(ByVal gameTime As GameTime)
			' Dark blue sky
			GraphicsDevice.Clear(New Color(0, 0, &H20))

			spriteBatch.Begin(SpriteSortMode.Immediate, Nothing, Nothing, Nothing, Nothing, Nothing, displayMatrix)

			' Render horizon line in green
			For i  = 0 To horizon.Length - 1
				Dim point1 As Vector2 = horizon(i)
				Dim point2 As Vector2 = horizon((i + 1) Mod horizon.Length)

				If (Not Single.IsNaN(point1.X)) AndAlso (Not Single.IsNaN(point2.X)) Then
					lineRenderer.Draw(spriteBatch, point1, point2, 3, Color.Green)
				End If
			Next i

			' Render horizon labels in green
			For i  = 0 To horizonLabels.Length - 1
				Dim textSize As Vector2 = labelFont.MeasureString(horizonLabels(i))
				Dim horizonIndex As Integer = i * horizon.Length \ horizonLabels.Length
				spriteBatch.DrawString(labelFont, horizonLabels(i), New Vector2(horizon(horizonIndex).X - textSize.X / 2, horizon(horizonIndex).Y - textSize.Y), Color.Green)
			Next i

			' Render ecliptic line in red
			For i  = 0 To ecliptic.Length - 1
				Dim point1 As Vector2 = ecliptic(i)
				Dim point2 As Vector2 = ecliptic((i + 1) Mod ecliptic.Length)
				lineRenderer.Draw(spriteBatch, point1, point2, 3, Color.Red)

				' And the hourly tick marks
				lineRenderer.Draw(spriteBatch, eclipticTicks(i, 0), eclipticTicks(i, 1), 3, Color.Red)
			Next i

			' Loop through constellations
			For Each constellation As Constellation In constellations.ConstellationList
				If Not Single.IsNaN(constellation.ScreenX) Then
					' Display constellation name
					Dim textSize As Vector2 = labelFont.MeasureString(constellation.LongName)
					spriteBatch.DrawString(labelFont, constellation.LongName, New Vector2(constellation.ScreenX - textSize.X / 2, constellation.ScreenY - textSize.Y / 2), Color.Gray)
				End If

				' Display constellation connectors
				If constellation.StarConnectors IsNot Nothing Then
					For Each starConnector As StarConnector In constellation.StarConnectors
						Dim point1 As New Vector2(CSng(starConnector.From.ScreenX), CSng(starConnector.From.ScreenY))
						Dim point2 As New Vector2(CSng(starConnector.To.ScreenX), CSng(starConnector.To.ScreenY))

						If (Not Single.IsNaN(point1.X)) AndAlso (Not Single.IsNaN(point2.X)) Then
							lineRenderer.Draw(spriteBatch, point1, point2, 1, Color.White)
						End If
					Next starConnector
				End If
			Next constellation

			' Now display the stars themselves
			For Each star As Star In visibleStars
				If Not Single.IsNaN(star.ScreenX) Then
					starDotRenderer.Draw(spriteBatch, New Vector2(star.ScreenX, star.ScreenY), Color.White)
				End If
			Next star

			' Finally the planets and Sun
			For Each systemBody As SolarSystemBody In systemBodies
				If Not Single.IsNaN(systemBody.ScreenX) Then
					planetDotRenderer.Draw(spriteBatch, New Vector2(systemBody.ScreenX, systemBody.ScreenY), systemBody.Color)
					Dim textSize As Vector2 = labelFont.MeasureString(systemBody.Name)
					spriteBatch.DrawString(labelFont, systemBody.Name, New Vector2(systemBody.ScreenX - textSize.X / 2, systemBody.ScreenY + 6), systemBody.Color)
				End If
			Next systemBody

			spriteBatch.End()
			MyBase.Draw(gameTime)
		End Sub
	End Class
End Namespace
