Namespace Petzold.Astronomy
	Public Structure Angle
		Private Const degreesToRadians As Double = Math.PI / 180
		Private Const radiansToDegrees As Double = 180 / Math.PI
        Private _radians As Double

        ' Basic properties
        Public Property Radians As Double
            Get
                Return _radians
            End Get
            Private Set(ByVal value As Double)
                _radians = value
            End Set
        End Property

        Public Property Degrees As Double
            Get
                Return radiansToDegrees * _radians
            End Get
            Private Set(ByVal value As Double)
                _radians = degreesToRadians * value
            End Set
        End Property

        ' Static properties of type Angle
        Public Shared ReadOnly Property Zero As Angle
            Get
                Return New Angle
            End Get
        End Property

        Public Shared ReadOnly Property Right As Angle
            Get
                Return Angle.FromDegrees(90)
            End Get
        End Property

        Public Shared ReadOnly Property Straight As Angle
            Get
                Return Angle.FromDegrees(180)
            End Get
        End Property

        Public Shared ReadOnly Property Circle As Angle
            Get
                Return Angle.FromDegrees(360)
            End Get
        End Property

        ' Other static creation properties
        Public Shared Function FromRadians(ByVal radians As Double) As Angle
            Dim angle As New Angle
            angle.Radians = radians
            Return angle
        End Function

        Public Shared Function FromDegrees(ByVal degrees As Double) As Angle
            Dim angle As New Angle
            angle.Degrees = degrees
            Return angle
        End Function

        Public Shared Function FromDegrees(ByVal degrees As Integer, ByVal minutes As Double) As Angle
            Return FromDegrees(degrees + minutes / 60)
        End Function

        Public Shared Function FromDegrees(ByVal degrees As Integer, ByVal minutes As Integer, ByVal seconds As Double) As Angle
            Return FromDegrees(degrees, minutes + seconds / 60)
        End Function

        Public Shared Function FromHours(ByVal hours As Double) As Angle
            Dim angle As New Angle
            angle.Degrees = 15 * hours
            Return angle
        End Function

        Public Shared Function FromHours(ByVal hours As Integer, ByVal minutes As Double) As Angle
            Return FromHours(hours + minutes / 60)
        End Function

        Public Shared Function FromHours(ByVal hours As Integer, ByVal minutes As Integer, ByVal seconds As Double) As Angle
            Return FromHours(hours, minutes + seconds / 60)
        End Function

        ' Trig functions
        Public Property Sine As Double
            Get
                Return Math.Sin(_radians)
            End Get
            Set(ByVal value As Double)
                _radians = Math.Asin(value)
            End Set
        End Property

        Public Property Cosine As Double
            Get
                Return Math.Cos(_radians)
            End Get
            Set(ByVal value As Double)
                _radians = Math.Acos(value)
            End Set
        End Property

        Public Property Tangent As Double
            Get
                Return Math.Tan(_radians)
            End Get
            Set(ByVal value As Double)
                _radians = Math.Atan(value)
            End Set
        End Property

        Public Shared Function ArcTangent(ByVal y As Double, ByVal x As Double) As Angle
            Return Angle.FromRadians(Math.Atan2(y, x))
        End Function

        ' Normalization
        Public Sub NormalizeAroundZero()
            Radians = Radians Mod 2 * Math.PI

            Do While Radians < -Math.PI
                Radians += 2 * Math.PI
            Loop

            Do While Radians > Math.PI
                Radians -= 2 * Math.PI
            Loop
        End Sub

        Public Sub NormalizePositive()
            Radians = Radians Mod 2 * Math.PI

            Do While Radians < 0
                Radians += 2 * Math.PI
            Loop

            Do While Radians > 2 * Math.PI
                Radians -= 2 * Math.PI
            Loop
        End Sub

        ' Operators and overrides
        Public Shared Operator -(ByVal angle As Angle) As Angle
            Return Angle.FromRadians(-angle.Radians)
        End Operator

        Public Shared Operator +(ByVal ang1 As Angle, ByVal ang2 As Angle) As Angle
            Return Angle.FromRadians(ang1.Radians + ang2.Radians)
        End Operator

        Public Shared Operator -(ByVal ang1 As Angle, ByVal ang2 As Angle) As Angle
            Return Angle.FromRadians(ang1.Radians - ang2.Radians)
        End Operator

        Public Shared Operator *(ByVal angle As Angle, ByVal multiplier As Double) As Angle
            Return Angle.FromRadians(angle.Radians * multiplier)
        End Operator

        Public Shared Operator *(ByVal multiplier As Double, ByVal angle As Angle) As Angle
            Return Angle.FromRadians(multiplier * angle.Radians)
        End Operator

        Public Shared Operator /(ByVal angle1 As Angle, ByVal angle2 As Angle) As Double
            Return angle1.Radians / angle2.Radians
        End Operator

        Public Shared Operator /(ByVal angle As Angle, ByVal divisor As Double) As Angle
            Return Angle.FromRadians(angle.Radians / divisor)
        End Operator

        Public Shared Operator <(ByVal angle1 As Angle, ByVal angle2 As Angle) As Boolean
            Return angle1.Radians < angle2.Radians
        End Operator

        Public Shared Operator >(ByVal angle1 As Angle, ByVal angle2 As Angle) As Boolean
            Return angle1.Radians > angle2.Radians
        End Operator

        Public Shared Operator <=(ByVal angle1 As Angle, ByVal angle2 As Angle) As Boolean
            Return angle1.Radians <= angle2.Radians
        End Operator

        Public Shared Operator >=(ByVal angle1 As Angle, ByVal angle2 As Angle) As Boolean
            Return angle1.Radians >= angle2.Radians
        End Operator

        Public Shared Operator =(ByVal angle1 As Angle, ByVal angle2 As Angle) As Boolean
            Return angle1.Radians = angle2.Radians
        End Operator

        Public Shared Operator <>(ByVal angle1 As Angle, ByVal angle2 As Angle) As Boolean
            Return angle1.Radians <> angle2.Radians
        End Operator

        Public Shared Narrowing Operator CType(ByVal angle As Angle) As Integer
            Return CInt(angle.Radians)
        End Operator

        Public Overloads Overrides Function Equals(ByVal obj As Object) As Boolean
            If Not (TypeOf obj Is Angle) Then
                Return False
            End If

            Return Equals(CType(obj, Angle))
        End Function

        Public Overloads Function Equals(ByVal that As Angle) As Boolean
            Return Me._radians.Equals(that._radians)
        End Function

        Public Overrides Function GetHashCode() As Integer
            Return _radians.GetHashCode
        End Function

		Public Overrides Function ToString() As String
			Return Degrees.ToString & ChrW(&H00B0).ToString
		End Function
	End Structure
End Namespace
