Namespace Petzold.Astronomy
	Public Structure HorizontalCoordinate
		Public Sub New(ByVal azimuth As Angle, ByVal altitude As Angle, ByVal tilt As Angle)
            Me.Azimuth = azimuth
			Me.Altitude = altitude
			Me.Tilt = tilt
		End Sub

		Public Sub New(ByVal azimuth As Angle, ByVal altitude As Angle)
			Me.New(azimuth, altitude, Angle.Zero)
		End Sub

		' Eastward from north
        Private _azimuth As Angle
        Public Property Azimuth As Angle
            Get
                Return _azimuth
            End Get
            Private Set(ByVal value As Angle)
                _azimuth = value
            End Set
        End Property

        Private _altitude As Angle
        Public Property Altitude As Angle
            Get
                Return _altitude
            End Get
            Private Set(ByVal value As Angle)
                _altitude = value
            End Set
        End Property

        Private _tilt As Angle
        Public Property Tilt As Angle
            Get
                Return _tilt
            End Get
            Private Set(ByVal value As Angle)
                _tilt = value
            End Set
        End Property

		Public Shared Function From(ByVal vector As Vector3) As HorizontalCoordinate
			Dim altitude As Angle = Angle.Zero
			altitude.Sine = vector.Z

			' Make X and Y components negative for zero azimuth at south
			'  with westward increasing values.
			' Then change x and y calculations likewise in ToVector method
			'  and remove adjustment in From(EquatorialCoordinates) method.
			Dim azimuth As Angle = Angle.ArcTangent(vector.X, vector.Y)

			Return New HorizontalCoordinate(azimuth, altitude)
		End Function

		Public Shared Function From(ByVal matrix As Matrix) As HorizontalCoordinate
			' Invert the matrix 
			matrix = Matrix.Invert(matrix)

			' Transform (0, 0, -1) -- the vector extending from the lens
			Dim zAxisTransformed As Vector3 = Vector3.Transform(-Vector3.UnitZ, matrix)

			' Get the horizontal coordinates
			Dim horzCoord As HorizontalCoordinate = From(zAxisTransformed)

			' Find the theoretical HorizontalCoordinate for the transformed +Y vector if the phone is upright
			Dim yUprightAltitude As Angle = Angle.Zero
			Dim yUprightAzimuth As Angle = Angle.Zero

			If horzCoord.Altitude.Degrees > 0 Then
				yUprightAltitude = Angle.Right - horzCoord.Altitude
				yUprightAzimuth = Angle.Straight + horzCoord.Azimuth
			Else
				yUprightAltitude = Angle.Right + horzCoord.Altitude
				yUprightAzimuth = horzCoord.Azimuth
			End If
			Dim yUprightVector As Vector3 = New HorizontalCoordinate(yUprightAzimuth, yUprightAltitude).ToVector

			' Find the real transformed +Y vector
			Dim yAxisTransformed As Vector3 = Vector3.Transform(Vector3.UnitY, matrix)

			' Get the angle between the upright +Y vector and the real transformed +Y vector
			Dim dotProduct As Double = Vector3.Dot(yUprightVector, yAxisTransformed)
			Dim crossProduct As Vector3 = Vector3.Cross(yUprightVector, yAxisTransformed)
			crossProduct.Normalize()

			' Sometimes dotProduct is slightly greater than 1, which 
			' raises an exception in the angleBetween calculation, so....
			dotProduct = Math.Min(dotProduct, 1)

			Dim angleBetween As Angle = Angle.FromRadians(Vector3.Dot(zAxisTransformed, crossProduct) * Math.Acos(dotProduct))
			horzCoord.Tilt = angleBetween

			Return horzCoord
		End Function

		Public Shared Function From(ByVal equatorialCoordinate As EquatorialCoordinate, ByVal geographicCoordinate As GeographicCoordinate, ByVal time As Time) As HorizontalCoordinate
			' Calculate hour angle
			Dim localHourAngle As Angle = time.GreenwichSiderealTime - geographicCoordinate.Longitude - equatorialCoordinate.RightAscension

			' Calculate azimuth
			Dim azimuth As Angle = Angle.ArcTangent(localHourAngle.Sine, localHourAngle.Cosine * geographicCoordinate.Latitude.Sine - equatorialCoordinate.Declination.Tangent * geographicCoordinate.Latitude.Cosine)

			' Adjustment for azimuth eastward from north
			azimuth += Angle.Straight

			azimuth.NormalizeAroundZero()

			' Calculate altitude
			Dim altitude As Angle = Angle.Zero
			altitude.Sine = geographicCoordinate.Latitude.Sine * equatorialCoordinate.Declination.Sine + geographicCoordinate.Latitude.Cosine * equatorialCoordinate.Declination.Cosine * localHourAngle.Cosine
			altitude.NormalizeAroundZero()

			Return New HorizontalCoordinate(azimuth, altitude)
		End Function

		Public Function ToVector() As Vector3
			Dim x As Double = Me.Altitude.Cosine * Me.Azimuth.Sine
			Dim y As Double = Me.Altitude.Cosine * Me.Azimuth.Cosine
			Dim z As Double = Me.Altitude.Sine

			Return New Vector3(CSng(x), CSng(y), CSng(z))
		End Function

		Public Overrides Function ToString() As String
			Return String.Format("Azi: {0} Alt: {1} Tilt: {2}", Me.Azimuth, Me.Altitude, Me.Tilt)
		End Function
	End Structure
End Namespace
