Namespace Petzold.Astronomy
	Public Structure Time
        Private _universalTime As Date

        ' Cached values
        Private _dynamicalTime As Date
        Private _julianEphemerisDay As Double
        Private _greenwichSiderealTime As Angle

        ' Static creation methods
        Public Shared Function GetNow() As Time
            Return CreateTime(Date.UtcNow)
        End Function

        Public Shared Function FromLocalTime(ByVal datetime As Date) As Time
            Return CreateTime(datetime.ToUniversalTime)
        End Function

        Public Shared Function FromUniversalTime(ByVal datetime As Date) As Time
            Return CreateTime(datetime)
        End Function

        ' There's obviously some number crunching involved when a Time value is first created,
        '  but it's better to do it now instead of for numerous celestial objects.
        Shared Function CreateTime(ByVal universalTime As Date) As Time
            Dim time As New Time
            time._universalTime = universalTime

            ' Calculate dynamical time = univesal time + delta T
            '      from Meeus, pg 78 for 2000 through 2100
            Dim century As Double = (universalTime.Year - 2000) / 100
            Dim dt As Double = 102 + 102 * century + 25.3 * century * century
            dt += 0.37 * (universalTime.Year - 2100)
            Dim deltat As TimeSpan = TimeSpan.FromSeconds(dt)
            time._dynamicalTime = universalTime.Add(deltat)

            ' Calculate Julian Ephemeris Day 
            ' Meeus, pgs 60-61.
            Dim Y As Integer = time._dynamicalTime.Year
            Dim M As Integer = time._dynamicalTime.Month
            Dim D As Double = time._dynamicalTime.Day + (time._dynamicalTime.Hour + (time._dynamicalTime.Minute + (time._dynamicalTime.Second + time._dynamicalTime.Millisecond / 1000.0) / 60.0) / 60.0) / 24.0
            If M = 1 OrElse M = 2 Then
                Y -= 1
                M += 12
            End If

            Dim A As Integer = Y \ 100
            Dim B As Integer = 2 - A + A \ 4

            time._julianEphemerisDay = Math.Floor(365.25 * (Y + 4716)) + Math.Floor(30.6001 * (M + 1)) + D + B - 1524.5

            ' Calculate Greenwich Sideral Time
            ' Meeus, pg 87
            Dim t As Double = 10 * time.Tau

            time._greenwichSiderealTime = Angle.FromDegrees(280.46061837 + 360.98564736629 * (time._julianEphemerisDay - 2451545.0) + 0.000387933 * t * t - t * t * t / 38710000)

            time._greenwichSiderealTime.NormalizePositive()

            Return time
        End Function

        ' Read-only properties
        Public ReadOnly Property LocalTime As Date
            Get
                Return _universalTime.ToLocalTime
            End Get
        End Property

        Public ReadOnly Property UniversalTime As Date
            Get
                Return _universalTime
            End Get
        End Property

        Public ReadOnly Property DynamicalTime As Date
            Get
                Return _dynamicalTime
            End Get
        End Property

        Public ReadOnly Property Tau As Double
            Get
                Return (JulianEphemerisDay - 2451545.0) / 365250
            End Get
        End Property

        Public ReadOnly Property JulianEphemerisDay As Double
            Get
                Return _julianEphemerisDay
            End Get
        End Property

        Public ReadOnly Property GreenwichSiderealTime As Angle
            Get
                Return _greenwichSiderealTime
            End Get
        End Property

        ' Operators and overrides
        Public Shared Operator =(ByVal time1 As Time, ByVal time2 As Time) As Boolean
            Return time1._universalTime = time2._universalTime
        End Operator

        Public Shared Operator <>(ByVal time1 As Time, ByVal time2 As Time) As Boolean
            Return time1._universalTime <> time2._universalTime
        End Operator

        Public Overloads Overrides Function Equals(ByVal obj As Object) As Boolean
            If Not (TypeOf obj Is Time) Then
                Return False
            End If

            Return Equals(CType(obj, Time))
        End Function

        Public Overloads Function Equals(ByVal that As Time) As Boolean
            Return Me._universalTime.Equals(that._universalTime)
        End Function

        Public Overrides Function GetHashCode() As Integer
            Return _universalTime.GetHashCode
        End Function
	End Structure
End Namespace
